

class BaseMixinClass:
    logger = None
    messages = None
    rsi_data = None
    database_manager = None
    max_ships = None
    ship_data_headers = None

    @staticmethod
    def mention_user(user):
        raise NotImplementedError

    def get_author_if_given(self, author):
        raise NotImplementedError

    @staticmethod
    def print_dict_table(*args, **kwargs):
        raise NotImplementedError

    @staticmethod
    def print_list_table(*args, **kwargs):
        raise NotImplementedError

    def split_data_and_get_messages(self, *args, **kwargs):
        raise NotImplementedError



from abc import ABC

from tabulate import tabulate

from .road_map_mixin import RoadMapMixin


class RsiMixin(RoadMapMixin, ABC):

    def get_ship_data_from_name(self, ship_name):
        ship_data = self.rsi_data.get_ship(ship_name)
        if ship_data is None:
            found_ships = self.rsi_data.get_ships_by_query(ship_name)
            if len(found_ships) == 1:
                ship_data = found_ships[0]
            else:
                ship_data = found_ships
        return ship_data

    @staticmethod
    def get_ship_price_message(ship):
        return "*%s*  **%s**, price:  %s (+ VAT)" % (ship["manufacturer_code"], ship["name"], ship["price"])

    def get_ship_prices_lines(self, found_ships):
        prices_lines = []
        for ship in found_ships:
            try:
                prices_lines.append(self.get_ship_price_message(ship))
            except KeyError:
                prices_lines.append(self.messages.ship_price_unknown % (ship["manufacturer_code"], ship["name"]))
        return prices_lines

    def iterate_ship_prices(self, query, author=None):
        found_ships = self.rsi_data.get_ships_by_query(query)
        if 0 < len(found_ships) < self.max_ships:
            price_lines = self.get_ship_prices_lines(found_ships)
            for message in self.split_data_and_get_messages(price_lines, lambda lines: "\n".join(lines)):
                yield message
        else:
            yield self.messages.ship_not_exists % self.get_author_if_given(author)

    def format_ship_data(self, ship):
        table = [[key, ship.get(key, "unknown")] for key in self.ship_data_headers]
        return "%s\n```%s```\n" % (self.rsi_data.base_url + ship['url'], tabulate(table))

    def compare_ships_data(self, ships):
        table = [[key] for key in self.ship_data_headers]
        for ship in ships:
            if ship:
                for row in table:
                    row.append(ship.get(row[0], "unknown"))
        return "\n```%s```\n" % tabulate(table)

    def split_compare_if_too_long(self, ships):
        return self.split_data_and_get_messages(ships, self.compare_ships_data)

    def iterate_ship_info(self, query, author=None):
        found_ship = self.get_ship_data_from_name(query)
        if isinstance(found_ship, list) and (1 < len(found_ship) < self.max_ships):
            for message in self.split_compare_if_too_long(found_ship):
                yield message
        elif isinstance(found_ship, dict):
            yield self.format_ship_data(found_ship)
        else:
            yield self.messages.ship_not_exists % self.get_author_if_given(author)

    def iterate_ships_comparison(self, query, author=None):
        names = query.split(",")
        found_ships = []
        for name in names:
            for ship_data in self.rsi_data.get_ships_by_query(name.strip()):
                found_ships.append(ship_data)
        if isinstance(found_ships, list) and len(found_ships) < self.max_ships:
            for message in self.split_compare_if_too_long(found_ships):
                yield message
        else:
            yield self.messages.ship_not_exists % self.get_author_if_given(author)

    def update_releases(self):
        current_releases = self.rsi_data.get_updated_versions()
        self.database_manager.update_versions(current_releases)
        return "PU Live: %s\nPTU: %s\n" % (current_releases.get('live'), current_releases.get('ptu'))



from abc import ABC

from tabulate import tabulate

from ..base_mixin import BaseMixinClass


class RoadMapMixin(BaseMixinClass, ABC):

    def get_no_road_map_data_found_message(self):
            self.logger.debug("No Roadmap data found...")
            return self.messages.road_map_not_found % tabulate(self.rsi_data.road_map.get_releases_and_categories(),
                                                               tablefmt="fancy_grid")

    @staticmethod
    def _data_found(data, find):
        return (not find) or (find.lower() in data.lower())

    @staticmethod
    def _get_road_map_table(value, key):
        return "```\n%s\n```" % tabulate(value, tablefmt="presto", headers=(key, "Task", "Done"))

    def _get_road_map_data_message(self, data, find=None):
        result = []
        if data:
            self.logger.debug("Showing Roadmap...")
            if isinstance(data, str):
                if self._data_found(data, find):
                    result = [self.print_dict_table(data, table_format="fancy_grid")]
            elif isinstance(data, dict):
                for key, value in data.items():
                    if self._data_found(key + str(value), find):
                        result += self.split_data_and_get_messages(value, self._get_road_map_table, key)
        if result:
            return result
        else:
            return [self.get_no_road_map_data_found_message()]

    def get_road_map_messages(self, args):
        self.logger.debug("Requested Roadmap.")
        if args.category and args.version:
            result = self.rsi_data.road_map.get_release_category_details(args.version, args.category)
            result = self._get_road_map_data_message(result)
        elif args.version:
            result = self.rsi_data.road_map.get_release_details(args.version)
            result = self._get_road_map_data_message(result, find=args.find)
        elif args.category:
            result = self.rsi_data.road_map.get_category_details(args.category)
            result = self._get_road_map_data_message(result, find=args.find)
        elif args.list:
            result = self.rsi_data.road_map.get_releases_and_categories()
            result = [self.print_dict_table(result, table_format="fancy_grid")]
        elif args.find:
            result = self.rsi_data.road_map.get()
            result = self._get_road_map_data_message(result, find=args.find)
        elif args.update:
            result = self.rsi_data.road_map.update_data()
            result = [self.messages.success] if result else [self.messages.something_went_wrong]
        else:
            result = self.rsi_data.road_map.get_releases()
            result = [self.print_dict_table(result, table_format="fancy_grid")]
        return result



from .data_classes import Item, Price


class Commodity(Item):
    name_prefix = "commodity"

    @property
    def illegal(self):
        return self.get('%s_illegal' % self.name_prefix)

    @property
    def legal(self):
        return not self.illegal

    @property
    def buy_prices(self):
        return self._prices['buy']

    @property
    def best_buy(self):
        return min(price.value for price in self.buy_prices)

    @property
    def best_deal(self):
        return self.best_sell - self.best_buy

    def iterate_all_routes(self):
        for buy in self.buy_prices:
            for sell in self.sell_prices:
                yield buy, sell

    def iterate_filtered_routes(self, start_location=None, end_location=None, avoid=()):

        def location_matches_query(price, start_end_query):
            return not start_end_query or start_end_query.lower() in price.location.full_string.lower()

        def not_avoiding_location(price):
            return all(avoid_location.lower() not in price.location.full_string.lower() for avoid_location in avoid)

        def location_matches_conditions(price, query):
            return location_matches_query(price, query) and not_avoiding_location(price)

        for buy, sell in self.iterate_all_routes():   # type: CommodityPrice, CommodityPrice
            if location_matches_conditions(buy, start_location) and location_matches_conditions(sell, end_location):
                yield buy, sell

    def get_routes(self, cargo=576, money=50000, start_location=None, end_location=None, avoid=(), max_routes=4):

        def duplicate_value_route():
            for item in routes_data:
                if buy.value == item['buy price'] and sell.value == item['sell price']:
                    if buy.location.short_string not in item['buy locations']:
                        item['buy locations'] += "\n%s" % buy.location.short_string
                    if sell.location.short_string not in item['sell locations']:
                        item['sell locations'] += "\n%s" % sell.location.short_string
                    return True

        routes_data = []
        best_income = 0.0
        for buy, sell in self.iterate_filtered_routes(start_location, end_location, avoid):

            if duplicate_value_route():
                continue

            bought_units = cargo * 100
            spent_money = bought_units * buy.value
            if spent_money > money:
                bought_units = int(money / buy.value)
                spent_money = money

            income = round((sell.value * bought_units) - spent_money, 2)
            if income > best_income:
                best_income = income

            routes_data.append({
                'invested money': spent_money,
                'income': income,
                'bought units': round(bought_units, 2),
                'buy price': buy.value,
                'buy locations': buy.location.short_string,
                'sell price': sell.value,
                'sell locations': sell.location.short_string
            })
        if routes_data:
            routes_data.sort(key=lambda item: item.get('income'), reverse=True)
            routes_data = routes_data[:max_routes]
            table = [[key] + [item[key] for item in routes_data] for key in routes_data[0].keys()]

            return {
                '%s_name' % self.name_prefix: self.name,
                'best_income': best_income,
                'routes': routes_data,
                'table': table
            }


class Resource(Item):
    name_prefix = "resource"

    @property
    def type(self):
        return self.get('resource_type')

    def get_prices_table(self):
        return [{
            'aUEC/unit': price.value,
            'Locations': price.location.short_string
        } for price in self.sell_prices]


class CommodityPrice(Price):
    name_suffix = "commodity"


class ResourcePrice(Price):
    name_suffix = "resource"



from . import data_classes
from . import trade_classes


def choose_item_class(items):
    item_keys = items[0].keys()
    if "location_name" in item_keys:
        return data_classes.Location
    elif "container_name" in item_keys:
        return data_classes.Container
    elif "commodity_name" in item_keys:
        return trade_classes.Commodity
    elif "price_commodity" in item_keys:
        return trade_classes.CommodityPrice
    elif "resource_name" in item_keys:
        return trade_classes.Resource
    elif "price_resource" in item_keys:
        return trade_classes.ResourcePrice


class DataStructure(dict):
    def __init__(self, items_list, **kwargs):
        item_class = choose_item_class(items_list)
        super().__init__({item['id']: item_class(item, **kwargs) for item in items_list})

    def get_list(self):
        return list(self.values())

    def match_exact_name(self, query):
        for item in self.values():
            if query.lower() == item.name.lower():
                return item

    def match_names(self, query):
        return [item.id for item in self.values() if query.lower() in item.name.lower()]

    def exclude_names(self, query):
        return [item.id for item in self.values() if query.lower() not in item.name.lower()]

    def match_and_exclude_names(self, match_query, exclude_query):
        return [item.id for item in self.values()
                if match_query.lower() in item.name.lower() and exclude_query.lower() not in item.name.lower()]

    def auto_match_names(self, match_query, exclude_query):
        if match_query and exclude_query:
            return self.match_and_exclude_names(match_query, exclude_query)
        elif match_query:
            return self.match_names(match_query)
        elif exclude_query:
            return self.exclude_names(exclude_query)



import os
import json


class BaseClient:

    @staticmethod
    def read_json(file_name):
        file_path = os.path.join(os.path.dirname(__file__), "trade_data", file_name + ".json")
        with open(file_path, 'r') as f:
            return json.loads(f.read())

    def get_containers(self):
        return self.read_json('containers')

    def get_locations(self):
        return self.read_json('locations')


class TradeClient(BaseClient):

    def get_commodities(self):
        return self.read_json('commodities')

    def get_prices(self):
        return self.read_json('commodity_prices')


class MiningClient(BaseClient):

    def get_resources(self):
        return self.read_json('resources')

    def get_prices(self):
        return self.read_json('resource_prices')



from pymongo import MongoClient

from base_astro_bot.utils import MyLogger
from base_astro_bot.trade.data_structure import DataStructure
from base_astro_bot.trade.data_rat_client import TradeClient, MiningClient


class PricesStructure:

    def __init__(self, log_file="trade_assistant.log", mongo_client=None,
                 trade_data_client=TradeClient, mining_data_client=MiningClient):
        self._log_file = log_file
        self.logger = MyLogger(log_file_name=self._log_file, logger_name="Trade Assistant logger", prefix="[TRADE]")
        self.trade_data_client = trade_data_client()
        self.mining_data_client = mining_data_client()
        self.mongo_db = mongo_client   # type: MongoClient
        self.celestial_bodies = None   # type: DataStructure
        self.locations = None   # type: DataStructure
        self.commodities = None   # type: DataStructure
        self.commodity_prices = None   # type: DataStructure
        self.resources = None   # type: DataStructure
        self.resource_prices = None   # type: DataStructure
        self.update_data()

    def _update_data_structure(self):
        self.celestial_bodies = DataStructure(self.trade_data_client.get_containers())
        self.locations = DataStructure(self.trade_data_client.get_locations(), parents=self.celestial_bodies)
        self.commodity_prices = DataStructure(self.trade_data_client.get_prices(), locations=self.locations)
        self.commodities = DataStructure(self.trade_data_client.get_commodities(), prices=self.commodity_prices)
        self.resource_prices = DataStructure(self.mining_data_client.get_prices(), locations=self.locations)
        self.resources = DataStructure(self.mining_data_client.get_resources(), prices=self.resource_prices)
        return all((self.celestial_bodies, self.locations,
                    self.commodities, self.commodity_prices,
                    self.resources, self.resource_prices))

    def _save_cache(self, database_name, **kwargs):
        database = self.mongo_db.get_database(database_name)
        for name, value in kwargs.items():
            database.drop_collection(name)
            collection = database.create_collection(name)
            collection.insert_many(value)

    def _read_from_cache(self):
        self.celestial_bodies = DataStructure([doc for doc in self.mongo_db.places.celestial_bodies.find()])
        self.locations = DataStructure([doc for doc in self.mongo_db.places.locations.find()],
                                       parents=self.celestial_bodies)
        self.commodity_prices = DataStructure([doc for doc in self.mongo_db.trade.prices.find()],
                                              locations=self.locations)
        self.commodities = DataStructure([doc for doc in self.mongo_db.trade.commodities.find()],
                                         prices=self.commodity_prices)
        self.resource_prices = DataStructure([doc for doc in self.mongo_db.mining.prices.find()],
                                             locations=self.locations)
        self.resources = DataStructure([doc for doc in self.mongo_db.mining.resources.find()],
                                       prices=self.resource_prices)

    def update_data(self):
        if self._update_data_structure():
            if self.mongo_db:
                self._save_cache("places",
                                 celestial_bodies=self.celestial_bodies.get_list(),
                                 locations=self.locations.get_list())
                self._save_cache("trade",
                                 commodities=self.commodities.get_list(),
                                 prices=self.commodity_prices.get_list())
                self._save_cache("mining",
                                 resources=self.resources.get_list(),
                                 prices=self.resource_prices.get_list())
            return True
        elif self.mongo_db:
            self._read_from_cache()


class TradeAssistant(PricesStructure):

    def get_trade_routes(self, cargo=576, money=50000, start_location=None, end_location=None, avoid=(),
                         allow_illegal=True, max_commodities=3):
        all_routes = []
        for commodity in self.commodities.values():
            commodity_routes = commodity.get_routes(cargo, money, start_location, end_location, avoid)
            if commodity_routes and (commodity.legal or allow_illegal):
                all_routes.append(commodity_routes)

        if all_routes:
            all_routes.sort(key=lambda item: item.get('best_income'), reverse=True)

            return [(route['commodity_name'], route['table']) for route in all_routes[:max_commodities]]

    def get_mining_prices(self, resource_name=None):
        if resource_name:
            resource = self.resources.match_exact_name(resource_name)
            if resource:
                return resource.get_prices_table()
        else:
            return [
                {
                    "Resource": resource.name,
                    "aUEC/unit": "%.3f" % resource.best_sell,
                    "Locations": "\n".join(resource.best_sell_locations)
                }
                for resource in self.resources.values() if resource.best_sell
            ]


if __name__ == '__main__':
    ta = TradeAssistant()
    for _commodity_name, _routes_table in ta.get_trade_routes():
        print(_routes_table)




class DataItem(dict):
    def __init__(self, item, parents=None, locations=None, prices=None):
        super().__init__(item)
        self._prices = self._get_prices(prices)
        self._locations = locations
        self._parents = self._get_parents(parents)
        self._best_sell = self._get_best_sell()

    @property
    def id(self):
        return self.get('id')

    def _get_parents(self, all_parents):
        return []

    def _get_prices(self, all_prices):
        return {}

    def _get_best_sell(self):
        return None


class Item(DataItem):
    name_prefix = ""

    def _get_prices(self, all_prices):
        prices = {'buy': [], 'sell': []}
        for price in all_prices.values():
            if price.commodity_id == self.id:
                prices[price.type].append(price)
        return prices

    @property
    def name(self):
        return self.get('%s_name' % self.name_prefix)

    @property
    def sell_prices(self):
        return self._prices['sell']

    def _get_best_sell(self):
        if self.sell_prices:
            return max(price.value for price in self.sell_prices)

    @property
    def best_sell(self):
        return self._best_sell

    @property
    def best_sell_locations(self):
        return [price.location.short_name for price in self.sell_prices if price.value == self.best_sell]


class Price(DataItem):
    name_suffix = ""

    @property
    def commodity_id(self):
        return self.get('price_%s' % self.name_suffix)

    @property
    def location(self):
        location_id = self.get('price_location')
        return self._locations.get(location_id)

    @property
    def type(self):
        return self.get('price_type')

    @property
    def value(self):
        return self.get('price_unit_price')


class Container(DataItem):

    @property
    def name(self):
        return self.get('container_name')


class Location(DataItem):

    def _get_parent(self, all_parents):
        item_parent_id = self.get('location_container')
        return all_parents.get(item_parent_id)

    def _get_parents(self, all_parents):
        parents = []
        parent = self._get_parent(all_parents)
        while parent:
            parents.append(parent)
            parent_id = parent.get('container_parent')
            parent = all_parents.get(parent_id)
        return parents

    @property
    def name(self):
        return self.get('location_name')

    @property
    def short_name(self):
        short_name = self.name.replace("Mining", "M.").replace("Research", "R.")
        short_name = short_name.replace(" Area", "A.").replace(" Outpost", "O.").replace(" Facility", "F.")
        if len(short_name) > 20:
            short_name = short_name[:20]
        return short_name

    @property
    def parent(self):
        return self._parents[0]

    @property
    def planet(self):
        try:
            return self._parents[-2]
        except IndexError:
            return None

    @property
    def system(self):
        return self._parents[-1]

    @property
    def parents_string(self):
        return ", ".join([parent.name for parent in self._parents])

    @property
    def short_string(self):
        return "%s @ %s" % (self.short_name, self.parent.name)

    @property
    def full_string(self):
        return "%s @ %s" % (self.name, self.parents_string)



from abc import ABC

from tabulate import tabulate

from ..base_mixin import BaseMixinClass


class TradeMixin(BaseMixinClass, ABC):
    trade = None

    def get_trade_messages(self, args):
        budget = 50000
        cargo = 576
        avoid = []
        if args.budget:
            budget = float(args.budget)
            if budget < 1:
                budget = 1
        if args.cargo:
            cargo = int(args.cargo)
            if cargo < 1:
                cargo = 1
        if args.avoid:
            avoid = [item.strip() for item in args.avoid.split(",")]

        allow_illegal = not args.legal

        arguments = (cargo, budget, args.start_location, args.end_location, avoid, allow_illegal)
        for commodity_name, routes_table in self.trade.get_trade_routes(*arguments):
            yield self.format_table(commodity_name, routes_table)

    def format_table(self, commodity_name, routes_table):
        table_string = tabulate(routes_table, tablefmt='presto')
        line_length = max(len(line) for line in table_string.splitlines())
        header_position = int(line_length * 0.3)
        header_string = " commodity      |%s%s\n%s\n" % (" " * header_position, commodity_name, "-" * line_length)
        table_string = "```%s%s```" % (header_string, table_string)
        if len(table_string) < 2000:
            return table_string
        return self.format_table(commodity_name, routes_table[:-1])

    def get_mining_messages(self, resource):
        return self.print_dict_table(self.trade.get_mining_prices(resource))

    def update_trade_data(self):
        if self.trade.update_data():
            return self.messages.success
        else:
            return self.messages.something_went_wrong



from googleapiclient.discovery import build
from httplib2 import Http
from oauth2client import file, client, tools

import settings
from ..utils import MyLogger
from ..database import DatabaseManager


class PricesStructure:
    spreadsheet_id = settings.TRADE_SPREADSHEET_ID
    commodities = settings.TRADE_COMMODITIES

    def __init__(
            self,
            release=settings.TRADE_SC_RELEASE,
            cells_range=settings.TRADE_SPREADSHEET_CELLS_RANGE,
            log_file="trade_assistant.log",
            database_manager=None
    ):
        self._log_file = log_file
        self.logger = MyLogger(log_file_name=self._log_file, logger_name="Trade Assistant logger", prefix="[TRADE]")
        self.release = release
        self.range_name = '%s!%s' % (release, cells_range)
        self.database = self.get_given_or_new_database_manager(database_manager)
        self.locations = None
        self.prices = {}
        self._value_type_row = None
        self._initiate_database()

    def get_given_or_new_database_manager(self, database_manager):
        if database_manager:
            return database_manager
        else:
            return DatabaseManager(log_file=self._log_file)

    @staticmethod
    def _fill_missing_blank_cells(values):
        full_length = None
        for row in values:
            if 'Buy' in row:
                full_length = len(row)
                break
        for row in values:
            too_short_by = full_length - len(row)
            row += [''] * too_short_by
        return values

    def _get_spreadsheet_rows(self):
        store = file.Storage('google_token.json')
        credentials = store.get()
        if not credentials or credentials.invalid:
            flow = client.flow_from_clientsecrets('google_credentials.json', settings.TRADE_GOOGLE_SCOPES)
            credentials = tools.run_flow(flow, store)
        service = build('sheets', 'v4', http=credentials.authorize(Http()))

        result = service.spreadsheets().values().get(spreadsheetId=self.spreadsheet_id,
                                                     range=self.range_name).execute()
        return self._fill_missing_blank_cells(result.get('values', []))

    @staticmethod
    def _populate_merged_cells(row):
        populated_row = []
        previous_item = None
        for item in row:
            if previous_item and item == '':
                populated_row.append(previous_item)
            else:
                populated_row.append(item)
                previous_item = item
        return populated_row

    def _build_locations(self, locations, celestial_bodies):
        self.locations = {
            location: celestial_bodies[index]
            for index, location in enumerate(locations) if location
        }

    def _row_contains_price(self, index):
        return 'Buy' in self._value_type_row[index] or 'Sell' in self._value_type_row[index]

    def _download_data_structure(self):
        values = self._get_spreadsheet_rows()
        locations, celestial_bodies, self._value_type_row = None, None, None
        if values:
            for row in values:
                if 'YELA' in row:
                    celestial_bodies = self._populate_merged_cells(row)
                elif 'Port Olisar' in row:
                    if celestial_bodies:
                        self._build_locations(row, celestial_bodies)
                    locations = self._populate_merged_cells(row)
                elif 'Profit' in row:
                    self._value_type_row = row
                elif locations and celestial_bodies and self._value_type_row and row[0].strip() in self.commodities:
                    item_name = row[0]
                    item_prices = {}
                    for index, location in enumerate(locations):
                        if row[index] and self._row_contains_price(index):
                            price = str(float(row[index].replace(",", ".")))
                            transaction = self._value_type_row[index]
                            if item_prices.get(transaction):
                                if item_prices[transaction].get(price):
                                    item_prices[transaction][price].append(location)
                                else:
                                    item_prices[transaction][price] = [location]
                            else:
                                item_prices[transaction] = {price: [location]}
                    self.prices[item_name] = item_prices

    def _initiate_database(self):
        stored_data = self.database.get_trade_data()
        if stored_data:
            self.locations, self.prices = stored_data
        else:
            self.update_database()

    def update_database(self):
            self._download_data_structure()
            self.database.save_trade_data(self.locations, self.prices)



import requests
import json

import settings


def handles_request_exception(func):
    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except requests.exceptions.RequestException:
            return None
    return func_wrapper


class BaseClient:
    base_url = "https://scm.oceandatarat.org:8443/scm-server/api/v1/"
    token = settings.SCM_TOKEN
    headers = {'authorization': settings.SCM_TOKEN}

    @handles_request_exception
    def get_request(self, endpoint):
        return requests.get(self.base_url + endpoint, headers=self.headers).json()

    @handles_request_exception
    def patch_request(self, endpoint, payload):
        return requests.patch(self.base_url + endpoint, json=payload, headers=self.headers)

    @handles_request_exception
    def post_request(self, endpoint, payload):
        return requests.post(self.base_url + endpoint, json=payload, headers=self.headers)

    def get_containers(self):
        return self.get_request('containers')

    def get_locations(self):
        return self.get_request('locations')


class TradeClient(BaseClient):

    def get_commodities(self):
        return self.get_request('commodities')

    def get_prices(self):
        """
        :return: [
                  {
                    "id": {},
                    "price_type": "buy/sell",
                    "price_location": "string",
                    "price_date": "2018-11-15",
                    "price_commodity": "string",
                    "price_unit_price": 0
                  }
                ]
        """
        return self.get_request('commodity_prices')

    def update_price(self, commodity_id, price, location_id, transaction_type="buy"):
        payload = {
                    "price_type": transaction_type,
                    "price_location": location_id,
                    "price_commodity": commodity_id,
                    "price_unit_price": price
                  }
        return self.patch_request('commodity_prices', payload)

    def add_transaction(self, commodity_id, price, units, location_id, transaction_type):
        payload = {
                    "transaction_type": transaction_type,
                    "transaction_location": location_id,
                    "transaction_commodity": commodity_id,
                    "transaction_units": units,
                    "transaction_unit_price": price
                  }
        return self.post_request('commodity_transactions', payload)


class MiningClient(BaseClient):

    def get_resources(self):
        return self.get_request('resources')

    def get_prices(self):
        """
        :return: [
                  {
                    "id": {},
                    "price_type": "buy/sell",
                    "price_location": "string",
                    "price_date": "2018-11-15",
                    "price_resource": "string",
                    "price_unit_price": 0
                  }
                ]
        """
        return self.get_request('resource_prices')

    def update_price(self, resource_id, price, location_id, transaction_type="buy"):
        payload = {
                    "price_type": transaction_type,
                    "price_location": location_id,
                    "price_resource": resource_id,
                    "price_unit_price": price
                  }
        return self.patch_request('resource_prices', payload)


def save_to_temp_file(text, file_name="temp.txt"):
    with open(file_name, 'w') as f:
        f.write(text)


def save_to_temp_json(structure, file_name="temp"):
    text = json.dumps(structure, indent=4)
    save_to_temp_file(text, file_name + ".json")


if __name__ == '__main__':
    cli = TradeClient()
    save_to_temp_json(cli.get_prices(), "trade_prices")
    # cli = MiningClient()
    # save_to_temp_json(cli.get_resources(), "mining_resources")



import threading
import time

from tabulate import tabulate
from pymongo import MongoClient
import pafy

from .utils import MyLogger
from .database import DatabaseManager
from .rsi_data import RsiDataParser, RsiMixin
from .trade import TradeAssistant, TradeMixin
from .fleet import FleetMixin
import settings


class BaseBot(RsiMixin, TradeMixin, FleetMixin):
    main_channel_id = settings.CHANNELS['main']
    messages = settings.messages
    max_characters = settings.MESSAGE_MAX_CHARACTERS
    max_ships = settings.MESSAGE_MAX_SHIPS
    ship_data_headers = settings.SHIP_DATA_HEADERS
    mongo_uri = settings.MONGO_CONNECTION_STRING

    def __init__(self):
        self.logger = MyLogger(log_file_name=settings.LOG_FILE, logger_name=settings.LOGGER_NAME, prefix="[BOT]")

        self.channel_main = self._get_channel_instance(self.main_channel_id)
        self.bot_user = self._get_bot_user()

        self.database_manager = DatabaseManager(log_file=settings.LOG_FILE)
        self.mongo = MongoClient(self.mongo_uri)

        self.rsi_data = RsiDataParser(7600, log_file=settings.LOG_FILE, database_manager=self.database_manager)
        self.report_ship_price_list = settings.REPORT_SHIP_PRICE_LIST

        self.monitoring_thread = threading.Thread(target=self.monitoring_procedure)
        self.monitoring_thread.start()

        self.trade = TradeAssistant(log_file=settings.LOG_FILE, mongo_client=self.mongo)

        self.help_message = self._get_help_message()

    def _get_channel_instance(self, channel_id):
        raise NotImplementedError

    def _get_help_message(self):
        raise NotImplementedError

    def _get_bot_user(self):
        raise NotImplementedError

    @staticmethod
    def mention_user(user):
        raise NotImplementedError

    @staticmethod
    def mention_channel(channel):
        raise NotImplementedError

    def get_author_if_given(self, author):
        if author:
            return self.mention_user(author)
        else:
            return ""

    @staticmethod
    def print_dict_table(items, table_format="presto"):
        return "```%s```" % tabulate(items, headers='keys', tablefmt=table_format)

    @staticmethod
    def print_list_table(items, table_format="presto"):
        return "```%s```" % tabulate(items, tablefmt=table_format)

    def split_data_and_get_messages(self, items, get_message_function, *args, **kwargs):
        message = get_message_function(items, *args, **kwargs)
        if len(message) < self.max_characters:
            messages = [message]
        else:
            half_length = int(len(items) * 0.5)
            messages = self.split_data_and_get_messages(items[:half_length], get_message_function, *args, **kwargs)
            messages += self.split_data_and_get_messages(items[half_length:], get_message_function, *args, **kwargs)
        return messages

    def report_ship_price(self):
        for ship_name, price_limit in self.report_ship_price_list:
            ship_data = self.rsi_data.get_ship(ship_name)
            if ship_data is None:
                self.channel_main.send_message(self.messages.ship_from_report_not_found % ship_name)
            else:
                current_price = float(ship_data["price"][1:])
                if current_price > price_limit:
                    self.channel_main.send_message(self.messages.ship_price_report % (ship_name, ship_data["price"]))
                    self.report_ship_price_list.remove((ship_name, price_limit))

    def monitor_current_releases(self):
        current_releases = self.rsi_data.get_updated_versions()
        new_version_released = self.database_manager.update_versions(current_releases)

        if new_version_released:
            new_release_message = self.update_releases()
            new_release_message = self.messages.new_version % new_release_message
            self.channel_main.send_message(new_release_message)

    def monitor_forum_threads(self):
        new_threads = self.rsi_data.get_forum_release_messages()
        if new_threads:
            self.channel_main.send_message(self.messages.new_version % "")
            for thread in new_threads:
                self.channel_main.send_message(thread)

    def monitor_youtube_channel(self):
        latest_video_url = pafy.get_channel("RobertsSpaceInd").uploads[0].watchv_url
        if self.database_manager.rsi_video_is_new(latest_video_url):
            self.channel_main.send_message(latest_video_url)

    def monitoring_procedure(self):
        while True:
            self.monitor_current_releases()
            self.monitor_forum_threads()
            self.monitor_youtube_channel()
            self.report_ship_price()
            time.sleep(300)



from abc import ABC
from operator import itemgetter

from ..base_mixin import BaseMixinClass


class FleetMixin(BaseMixinClass, ABC):

    def get_ship_data_from_name(self, ship_name):
        raise NotImplementedError

    def clear_member_fleet(self, author):
        self.database_manager.update_member_ships([], author)
        return self.database_manager.get_ships_by_member_name(author.username)

    def get_fleet_tables(self, args):
        if args.all_ships:
            ships = self.database_manager.get_all_ships_dicts()
        else:
            ships = self.database_manager.get_ships_summary()

        if args.filter:
            filters = args.filter.split(",")
            for item in filters:
                key, expected_value = item.split("=")
                ships = [ship for ship in ships if expected_value.lower() in str(ship[key]).lower()]

        if args.order_by:
            columns = args.order_by.split(",")
            columns.reverse()
        else:
            columns = ["name", "manufacturer"]
        for column in columns:
            ships = sorted(ships, key=itemgetter(column))
        if args.descending:
            ships.reverse()

        return self.split_data_and_get_messages(ships, self.print_dict_table)

    def get_ship_for_member(self, ship):
        ship_name = ship.replace("lti", "").strip()
        ship_data = self.get_ship_data_from_name(ship_name)
        if ship_data and isinstance(ship_data, dict):
            ship_data['lti'] = ship[-3:].lower() == "lti"
            return ship_data

    def add_member_ship(self, ship_query, author):
        ship_data = self.get_ship_for_member(ship_query)
        if ship_data:
            self.database_manager.add_one_ship(ship_data, author)
            for message in self.iterate_updated_ships_messages(author):
                yield message
        else:
            yield self.messages.ship_not_exists % self.mention_user(author)

    def remove_member_ship(self, ship_query, author):
        ship_data = self.get_ship_for_member(ship_query)
        if ship_data:
            if self.database_manager.remove_one_ship(ship_data, author):
                for message in self.iterate_updated_ships_messages(author):
                    yield message
            else:
                yield self.messages.member_ship_not_found % self.mention_user(author)
        else:
            yield self.messages.ship_not_exists % self.mention_user(author)

    def iterate_invalid_ships_messages(self, author, invalid_ships):
        yield self.messages.member_ships_invalid % (self.mention_user(author))
        for message in self.split_data_and_get_messages(invalid_ships, self.print_dict_table, table_format="rst"):
            yield message

    def iterate_updated_ships_messages(self, author):
        yield self.messages.member_ships_modified % (self.mention_user(author))
        for message in self.get_member_fleet(author.username):
            yield message

    def get_member_fleet(self, member_name):
        ships = self.database_manager.get_ships_dicts_by_member_name(member_name)
        if ships:
            for message in self.split_data_and_get_messages(ships, self.print_dict_table, table_format="rst"):
                yield message
        else:
            yield self.messages.member_not_found




TEST_SHIPS = [
  {
    "manufacturer": "Drake",
    "name": "Herald",
    "lti": True,
    "warbond": False,
    "package_id": "9635433",
    "pledge": "Standalone Ship - MPUV Personnel - LTI",
    "pledge_date": "August 30, 2016",
    "cost": "$85.00 USD"
  },
  {
    "manufacturer": "Drake",
    "name": "Cutlass Black",
    "lti": True,
    "warbond": True,
    "package_id": "12616268",
    "pledge": "Standalone Ship - Aopoa Nox Kue LTI  - Warbond",
    "pledge_date": "August 23, 2017",
    "cost": "$90.00 USD"
  },
  {
    "manufacturer": "RSI",
    "name": "Constellation Taurus",
    "lti": True,
    "warbond": False,
    "package_id": "12620007",
    "pledge": "Standalone Ship - MPUV Cargo - LTI",
    "pledge_date": "August 24, 2017",
    "cost": "$150.00 USD"
  }
]
